/* arcfour.c
 *
 * This implements the Arcfour stream cipher with 128 bit keys. 
 *
 * The Arcfour cipher is believed to be compatible with the RC4 cipher. 
 * RC4 is a registered trademark of RSA Data Security Inc.
 */
/*
 * This file has been modified for libcftp - cftp protocol implementation.
 *
 * $Id: arcfour.c,v 1.1.1.1 2000/04/22 10:15:52 malekith Exp $
 */

#include "config.h"
#include "utils.h"
#include "cipher.h"

#include <string.h>

struct arcfour_ctx {
  struct cipher parent;

  byte S[256];
  byte i, j;
};


#define SWAP(a,b) do { int _t = a; a = b; b = _t; } while(0)

static void arcfour_set_key(struct arcfour_ctx *ctx, const byte *key, u32 len)
{
  register byte j; /* Depends on the eight-bitness of these variables. */
  unsigned i;
  u32 k;

  /* Initialize context */
  i = 0;
  do ctx->S[i] = i; while (++i < 256);

  /* Expand key */
  i = j = k = 0;
  do {
    j += ctx->S[i] + key[k];
    SWAP(ctx->S[i], ctx->S[j]);
    k = (k+1) % len; /* Repeat key if needed */
  } while(++i < 256);
  
  ctx->i = ctx->j = 0;
}

static void arcfour_crypt(struct arcfour_ctx *ctx, byte *p, u32 len)
{
  register byte i, j;

  i = ctx->i; j = ctx->j;
  while(len--)
    {
      i++; /* i &= 0xff; */
      j += ctx->S[i]; /* j &= 0xff; */
      SWAP(ctx->S[i], ctx->S[j]);
      *p++ ^= ctx->S[ (ctx->S[i] + ctx->S[j]) & 0xff ];
    }
  ctx->i = i; ctx->j = j;
}

static void dfree(struct cipher *self)
{
	xfree0(self, sizeof(struct arcfour_ctx));
}

struct cipher *arcfour_create()
{
	struct cipher *r;
	r = (struct cipher *)xmalloc(sizeof(struct arcfour_ctx));
	r->init = (init_cipher_func)arcfour_set_key;
	r->cipher = (cipher_func)arcfour_crypt;
	r->decipher = (cipher_func)arcfour_crypt;
	r->free = dfree;
	
	r->keylen = 256;
	r->blocklen = 1;
	r->name = "arcfour";
	r->disabled = 0;
	
	return r;
}
