/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: clib1.c,v 1.9 2000/05/05 22:20:26 malekith Exp $
 */

#include "config.h"
#include "clib_int.h"
#include "rsa.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>

#ifdef WINDA
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/time.h>
#endif

char *home_dir;

static int get_port_def(const char *serv, int def)
{
	struct servent *s;
	s = getservbyname(serv, "tcp");
	return s ? ntohs(s->s_port) : def;
}

static void save_random_seed(void)
{
	byte rand_seed[512];
	FILE *f;
	char *fn;
	
	fn = xprintf("%s/random_seed", home_dir);
	
	random_get(rand_seed, sizeof(rand_seed));
	
	if ((f = fopen(fn, "wb")) != 0) {
		fwrite(rand_seed, 1, sizeof(rand_seed), f);
		fclose(f);
	}
	xfree(fn);
}

void clib_init(const char *etc_dir)
{
	byte rand_seed[512];
	char *fn;
	FILE *f;
	
	home_dir = xstrdup(etc_dir);
	crypto_init();
	fn = xprintf("%s/random_seed", home_dir);
	
	if ((f = fopen(fn, "rb")) != 0) {
		fread(rand_seed, 1, sizeof(rand_seed), f);
		fclose(f);
	}
	xfree(fn);
	
	random_add(rand_seed, sizeof(rand_seed));
	save_random_seed();
#ifdef WINDA
	init_wsa();
#endif
}

void clib_close(void)
{
	save_random_seed();
#ifdef WINDA
	kill_wsa();
#endif
}

int clib_open1(CTX * ctx, const char *host, int port)
{
	struct sockaddr_in sin;
#ifdef IPv6
	struct sockaddr_in6 sin6;
#endif
	int r;

	if (!port)
		port = get_port_def("cftp", DEF_PORT);
	ctx->host = xstrdup(host);

#ifdef IPv6
	if (strchr(host, ':')) {
		inet_pton(AF_INET6, host, sin6.sin6_addr.s6_addr);
		inet_pton(AF_INET6, host, ctx->u->target_addr);
		ctx->u->v6 = 1;
	} else
#endif
	if (isdigit(*host)) {
		*(u32 *) & sin.sin_addr = inet_addr(host);
		*(u32 *) ctx->u->target_addr = inet_addr(host);
	} else {
		struct hostent *h;

		h = gethostbyname(host);
		if (!h) {
			set_sock_error(ctx, host);
			closesocket(ctx->sock);
			return -1;
		}
#ifdef IPv6
		if (h->h_addrtype == AF_INET6) {
			memcpy(sin6.sin6_addr.s6_addr, h->h_addr_list[0],
			       16);
			memcpy(ctx->u->target_addr, h->h_addr_list[0], 16);
			ctx->u->v6 = 1;
		} else
#endif
		{
			memcpy(&sin.sin_addr, h->h_addr_list[0], 4);
			memcpy(ctx->u->target_addr, h->h_addr_list[0], 4);
		}
	}

#ifdef IPv6
	if (ctx->u->v6) {
		ctx->sock = socket(PF_INET6, SOCK_STREAM, 0);
		if (ctx->sock == INVALID_SOCKET)
			return set_sock_error(ctx, "socket6");
		sin6.sin6_family = AF_INET6;
		sin6.sin6_port = htons(port);
		r = connect(ctx->sock, (struct sockaddr *) &sin6,
			    sizeof(sin6));
	} else
#endif
	{
		ctx->sock = socket(PF_INET, SOCK_STREAM, 0);
		if (ctx->sock == INVALID_SOCKET)
			return set_sock_error(ctx, "socket");
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);
		r = connect(ctx->sock, (struct sockaddr *) &sin,
			    sizeof(sin));
	}

	if (r < 0) {
		set_sock_error(ctx, "couldn't connect");
		closesocket(ctx->sock);
		return -1;
	}

	return 0;
}


static int check_key(CTX *ctx, struct rsa_key * pub)
{
	char line[5000];
	FILE *f;
	char *key = 0, *fn;
	
	ctx->u->add_nl = 0;

	fn = xprintf("%s/known_hosts", home_dir);
	f = fopen(fn, "r");
	xfree(fn);
	
	if (!f)
		return 1; /* host not found */

	while (fgets(line, sizeof(line), f)) {
		char *p, *p2;

		ctx->u->add_nl = !strchr(line, '\n');
		p = line;
		while (isspace(*p))
			p++;

		if (*p == '#' || *p == '\n' || *p == '\r' || *p == 0)
			continue;

		for (p2 = p; *p2 && !isspace(*p2);)
			p2++;
		if (!*p2)
			continue;
		*p2++ = 0;

		if (xstrcasecmp(ctx->host, p) == 0) {
			key = p2;
			break;
		}
	}
	fclose(f);
	if (key)
		while (*key && isspace(*key))
			key++;
	if (!key || !*key)
		return 1;
	if (atoi(key) != (int)pub->bits)
		return 2;
	{
		mpz_t d;
		char *s;

		s = strchr(key, ' ');
		while (isspace(*s))
			s++;
		key = strchr(s, ' ');
		if (!key)
			return 2;
		*key++ = 0;
		mpz_init(d);
		mpz_set_str(d, s, 16);
		if (mpz_cmp(d, pub->e) != 0)
			return 2;
		mpz_set_str(d, key, 16);
		if (mpz_cmp(d, pub->n) != 0)
			return 2;
	}
	return 0;
}

static int set_pkey(CTX *ctx, struct pkt_key *pk, const char *cipher)
{
	int i;
	struct cipher2 *c2;
	
	for (i = 0; i < 32; i++) {
		pk->server_key[i] = rnd();
		pk->client_key[i] = rnd();
	}
	pk->server_key[0] |= 0x80;
	pk->client_key[0] |= 0x80;

	memset(pk->cipher_id, 0, sizeof(pk->cipher_id));
	strncpy(pk->cipher_id, cipher, sizeof(pk->cipher_id));
	pk->zero = 0;

	c2 = find_cipher(pk->cipher_id);

	if (!c2) {
		set_error(ctx,
			  xprintf("cipher %s is not available",
				  pk->cipher_id)); 
		return -1;
	}

	if (c2->disabled || !c2->creator) {
		set_error(ctx, xprintf("cipher %s is client "
				       "side disabled: %s", c2->name,
				       c2->disabled));
		return -1;
	}
	ctx->u->e = c2->creator();
	ctx->u->d = c2->creator();
	pk->used_key_len =
	    ctx->u->e->keylen <
	    (int) sizeof(pk->server_key) ? ctx->u->e->keylen : 
	    	  sizeof(pk->server_key);
	ctx->u->e->init(ctx->u->e, pk->client_key, pk->used_key_len);
	ctx->u->d->init(ctx->u->d, pk->server_key, pk->used_key_len);
	
	ctx->u->c2 = c2;
	ctx->u->ukl = pk->used_key_len;
	
	return 0;
}

/* after this ctx->u->key is encrypted and ready to send, ctx->u->[ed] are
 * created and initialised with proper keys */
static int rsa_stuff(CTX * ctx, const char *cipher)
{
	struct packet *p;
	struct pkt_key pk;
	struct rsa_key pub;
	int err;
	mpz_t key;

	p = recv_packet_x(ctx, PKT_RSAKEY);
	if (!p)
		return -1;
	err = rsa_read_pub_key(&pub, p->data);
	free_packet(p);

	if (err) {
		rsa_destroy_key(&pub);
		set_error(ctx, xprintf("bad rsa key from server"));
		return -2;
	}

	err = set_pkey(ctx, &pk, cipher);

	if (err) {
		rsa_destroy_key(&pub);
		return -3;
	}

	xmpz_from_byte(key, (byte*)&pk, sizeof(pk));
	memset(&pk, 0, sizeof(pk));
	rsa_encrypt(key, key, &pub);
	ctx->u->key = mpz_get_str(0, 16, key);

	err = check_key(ctx, &pub);

	if (err == 1) {
		char *s1, *s2;
		s1 = mpz_get_str(0, 16, pub.e);
		s2 = mpz_get_str(0, 16, pub.n);
		ctx->u->hostline = xprintf("%s%s %d %s %s\n",
					   ctx->u->add_nl ? "\n" : "",
					   ctx->host, pub.bits, s1, s2);
		xfree(s1);
		xfree(s2);
	}

	rsa_destroy_key(&pub);

	return err;
}

/* retval, if < 0 -- cannot continue
 *   0 - ok, host found in ~/known_hosts
 *   1 - new host, ask user whatever to continue
 *   2 - ALERT! host id changed !
 *  -1 - problems with CFTP-blah header (version_agreement())
 *  -2 - problems with format of server RSA key
 *  -3 - problems with cipher (client side)
 */
int clib_open2(CTX *ctx, const char *cipher)
{
	int r;
	
	r = version_agreement(ctx);
	if (!r)
		r = rsa_stuff(ctx, cipher);
	if (r < 0) {
		closesocket(ctx->sock);
		ctx->sock = INVALID_SOCKET;
	}
	return r;
}

static int send_cookie(CTX *ctx)
{
	byte cookie[32];
	struct packet *p;
	int r = 0;
	
	random_get(cookie, 32);
	send_packet(ctx, PKT_COOKIE, cookie, 32);
	p = recv_packet_x(ctx, PKT_COOKIE);
	if (!p)
		r = -1;
	else  if (p->size != 32 || memcmp(p->data, cookie, 32) != 0)
		r = set_error(ctx, xprintf("bad cookie from server -- "
			"possibly broken encryption method/key"));
	free_packet(p);
	return r;
}


int clib_open3(CTX *ctx)
{
	struct packet *p;
	struct pkt_key_rsp *r;
	
	/* check whatever we should add host to ~/known_hosts */
	if (ctx->u->hostline) {
		FILE *f;
		char *fn;
		
		fn = xprintf("%s/known_hosts", home_dir);
		f = fopen(fn, "a+");
		xfree(fn);
		if (f) {
			fputs(ctx->u->hostline, f);
			fclose(f);
		}
		xfree(ctx->u->hostline);
		ctx->u->hostline = 0;
	}
	
	if (send_packet(ctx, PKT_KEY, ctx->u->key, strlen(ctx->u->key)))
		return -1;
	xfree0(ctx->u->key, strlen(ctx->u->key));
	ctx->u->key = 0;
	
	p = recv_packet_x(ctx, PKT_KEY_RSP);
	if (!p) 
		return -1;
	r = (struct pkt_key_rsp *)p->data;
	
	if (r->accepted)
		set_error(ctx, xprintf("server accepted key (%s)", r->msg));
	else
		return set_error(ctx, xprintf("server has rejected key: %s", r->msg));
		
	ctx->u->encrypting = 1;
	
	/* check the connection (encryption) before sending passwords */	
	if (send_cookie(ctx))
		return -2;
	return 0;
}

int clib_open4(CTX *ctx, const char *user, const char *pass)
{
	int ul,pl,r;
	char *data;
	struct packet *p;
	
	ul = strlen(user);
	pl = strlen(pass);
	
	data = xmalloc(ul + 1 + pl + 1);
	memcpy(data, user, ul);
	data[ul] = 0;
	memcpy(data + ul + 1, pass, pl);
	data[ul + 1 + pl] = 0;
	if (send_packet(ctx, PKT_AUTH, data, ul + 1 + pl + 1))
		return -1;
	
	p = recv_packet(ctx);
	switch (p->id) {
	case PKT_ACCESS_DENIED:
		ctx->message = xstrdup(p->data);
		r = -1;
		break;
	case PKT_MSG:
		ctx->message = xstrdup(p->data);
		r = 0;
		break;
	default:
		ctx->message = xstrdup("unexpected packet during auth");
		r = -1;
		break;
	}
	free_packet(p);
	
	return r;
}

struct pkt_response *get_response_x(CTX * ctx, ... );
void free_resp(struct pkt_response *r);

/* ughh... what an ugly long pice of code ... i'm too lazy to break it ;-( */
int clib_forward(CTX * ctx, int port, const char *host, int local)
{
	struct pkt_command *cmd;
	struct cmd_forward *fwd;
	struct pkt_response *r;
	struct rsp_forward *rf;
	struct packet *p;
	int n, err;
	struct sockaddr_in sin;
#ifdef IPv6
	struct sockaddr_in6 sin6;
#endif
	CTX *nctx;
	SOCKET sock_local, sock_remote;
	byte client_key[32], server_key[32];

#ifdef IPv6
	if (ctx->u->v6) {
		sock_local = socket(PF_INET6, SOCK_STREAM, 0);
		sin6.sin6_port = htons(local);
		sin6.sin6_family = AF_INET6;
		sin6.sin6_addr = in6addr_any;
		if (sock_local != INVALID_SOCKET)
			err = bind(sock_local, (struct sockaddr *) &sin6,
				 sizeof(sin6));
	} else
#endif
	{
		sock_local = socket(PF_INET, SOCK_STREAM, 0);
		sin.sin_port = htons(local);
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
		if (sock_local != INVALID_SOCKET)
			err = bind(sock_local, (struct sockaddr *) &sin,
				 sizeof(sin));
	}

	if (sock_local == INVALID_SOCKET) {
		set_sock_error(ctx, "socket");
		ctx->u->dead = 0;
		return -2;
	}

	if (err < 0) {
		set_sock_error(ctx, "bind");
		closesocket(sock_local);
		ctx->u->dead = 0;
		return -2;
	}

	random_get(client_key, 32);
	random_get(server_key, 32);
	/* set first bit of keys, this isn't strictly required (no RSA 
	 * encryption), but ... */
	client_key[0] |= 0x80;
	server_key[0] |= 0x80;

	/*
	   n = sizeof(struct pkt_command) - 1 + sizeof(struct cmd_forward) - 1 
	   + strlen(host) + 1;
	 */
	n = 4 + 80 + strlen(host) + 1;

	cmd = (struct pkt_command *) xmalloc(n);
	fwd = (struct cmd_forward *) cmd->data;
	cmd->no = h2n(CMD_FORWARD);
	fwd->flags = 0;
	fwd->port = h2n(port);
	memcpy(fwd->client_key, client_key, 32);
	memcpy(fwd->server_key, server_key, 32);
	strcpy(fwd->host, host);
	send_packet(ctx, PKT_COMMAND, (byte *) cmd, n);
	xfree0(cmd, n);

	r = get_response_x(ctx, RSP_FORWARD, RSP_STATUS_ERR, 0);
	if (!r) {
		closesocket(sock_local);
		return -1;
	}

	if (r->no != RSP_FORWARD) {
		closesocket(sock_local);
		set_error(ctx,
			  xprintf("fwd -> %s:%d: %s", host, port,
				  r->data));
		free_resp(r);
		return 1;
	}

	rf = (struct rsp_forward *) r->data;

#ifdef IPv6
	if (ctx->u->v6) {
		sock_remote = socket(PF_INET6, SOCK_STREAM, 0);
		sleep(1);
		memcpy(sin6.sin6_addr.s6_addr, ctx->u->target_addr, 16);
		sin.sin_port = htons(n2h(rf->port));
		sin.sin_family = AF_INET6;
		if (sock_remote != INVALID_SOCKET)
			err = connect(sock_remote, (struct sockaddr *) &sin6,
				    sizeof(sin6));
	} else
#endif
	{
		sock_remote = socket(PF_INET, SOCK_STREAM, 0);
		sleep(1);
		memcpy(&sin.sin_addr, ctx->u->target_addr, 4);
		sin.sin_port = htons(n2h(rf->port));
		sin.sin_family = AF_INET;
		if (sock_remote != INVALID_SOCKET)
			err = connect(sock_remote, (struct sockaddr *) &sin,
				    sizeof(sin));
	}

	if (sock_remote == INVALID_SOCKET) {
		set_sock_error(ctx, "socket");
		closesocket(sock_local);
		ctx->u->dead = 0;
		return -3;
	}
	if (err < 0) {
		set_sock_error(ctx, "connect");
		ctx->u->dead = 0;
		closesocket(sock_local);
		closesocket(sock_remote);
		return -3;
	}

	nctx = clib_alloc();
	nctx->sock = sock_remote;
	nctx->u->fwd_local = sock_local;

	nctx->u->e = ctx->u->c2->creator();
	nctx->u->d = ctx->u->c2->creator();
	nctx->u->e->init(nctx->u->e, client_key, ctx->u->ukl);
	nctx->u->d->init(nctx->u->d, server_key, ctx->u->ukl);
	nctx->u->encrypting = 1;
	memset(client_key, 0, 32);
	memset(server_key, 0, 32);
	ctx->u->fwd_ctx = nctx;

	random_get(client_key, 32);
	send_packet(nctx, PKT_COOKIE, client_key, 32);
	p = recv_packet_x(nctx, PKT_COOKIE);
	if (!p || memcmp(p->data, client_key, 32) != 0) {
		free_packet(p);
		set_error(ctx, xprintf("cookie test failed"));
		clib_free(nctx);
		return -4;
	}

	free_packet(p);
	return 0;
}

static int fwd_loop(CTX * ctx, SOCKET sock)
{
	for (;;) {
		fd_set rd;

		FD_ZERO(&rd);
		FD_SET(sock, &rd);
		FD_SET(ctx->sock, &rd);

		select(sock < ctx->sock ? ctx->sock + 1 : sock + 1, &rd, 0,
		       0, 0);

		if (FD_ISSET(sock, &rd)) {
			byte buf[CHUNK_SIZE];
			int n;

			n = recv(sock, buf, sizeof(buf), 0);
			if (n == -1)
				return set_sock_error(ctx, "recv[local]");

			if (n == 0) {
				closesocket(sock);
				if (send_packet(ctx, PKT_FORWARD_EOF, 0,
				     0)) return -1;
				else
					return 0;
			}
			if (send_packet(ctx, PKT_FORWARD_DATA, buf, n))
				return -1;
		}

		if (FD_ISSET(ctx->sock, &rd)) {
			struct packet *p;
			p = recv_packet(ctx);
			if (!p)
				return -1;
			switch (p->id) {
			case PKT_FORWARD_DATA:
				if (write_l(sock, p->data, p->size) == -1) {
					set_sock_error(ctx, "send[local]");
					return -1;
				}
				break;
			case PKT_FORWARD_EOF:
				closesocket(sock);
			 	return 0;
			case PKT_DISCONNECT:
				set_error(ctx, xprintf("server requested "
						       "disconnection"));
				return -1;
			default:
				proto_err(ctx, "invalid packet");
				return -1;
			}
			free_packet(p);
		}
	}
}

char *clib_forward_forked(CTX *ctx)
{
	ctx = ctx->u->fwd_ctx;
	
	listen(ctx->u->fwd_local, 1);
	
	for (;;) {
		SOCKET fd;
#if 0
	/* i'm not sure whatever this is good idea ... */
		struct sockaddr_in sin;
		socklen_t len;
		
		len = sizeof(sin);
		fd = accept(ctx->u->fwd_local, (struct sockaddr*)&sin, &len);
		
		if (sin.sin_addr.s_addr != INADDR_LOOPBACK) {
			close(fd);
			continue;
		}
#endif

		fd = accept(ctx->u->fwd_local, 0, 0);
		if (fd == INVALID_SOCKET) {
			sleep(5);
			continue;
		}
		send_packet(ctx, PKT_FORWARD_CONNECT, 0, 0);
		if (fwd_loop(ctx, fd)) {
			char *r = 0;
			if (ctx->error)
				r = ctx->error;
			ctx->error = 0;
			clib_free(ctx);
			return r ? r : xstrdup("fwd_loop failed");
		}
	}
}

int clib_clr_error(CTX *ctx)
{
	if (ctx->u->dead)
		return -1;
	xfree(ctx->error);
	ctx->error = 0;
	return 0;
}

